/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "quakedef.h"

int	r_dlightframecount;


/*
==================
R_AnimateLight
==================
*/
void R_AnimateLight (void)
{
    int			j, k;
    float		l;
    int			flight;
    int			clight;
    float		lerpfrac;
    float		backlerp;

    // light animations
    // 'm' is normal light, 'a' is no light, 'z' is double bright
    flight = (int) floor (cl.time * 10);
    clight = (int) ceil (cl.time * 10);
    lerpfrac = (cl.time * 10) - flight;
    backlerp = 1.0f - lerpfrac;

    for (j = 0; j < MAX_LIGHTSTYLES; j++)
    {
        if (!cl_lightstyle[j].length)
        {
            // was 256, changed to 264 for consistency
            d_lightstylevalue[j] = 264;
            continue;
        }
        else if (cl_lightstyle[j].length == 1)
        {
            // single length style so don't bother interpolating
            d_lightstylevalue[j] = 22 * (cl_lightstyle[j].map[0] - 'a');
            continue;
        }

        // interpolate animating light
        // frame just gone
        k = flight % cl_lightstyle[j].length;
        k = cl_lightstyle[j].map[k] - 'a';
        l = (float) (k * 22) * backlerp;

        // upcoming frame
        k = clight % cl_lightstyle[j].length;
        k = cl_lightstyle[j].map[k] - 'a';
        l += (float) (k * 22) * lerpfrac;

        d_lightstylevalue[j] = (int)l;
    }
}

/*
=============================================================================

DYNAMIC LIGHTS BLEND RENDERING

=============================================================================
*/

void AddLightBlend (float r, float g, float b, float a2)
{
    float	a;

    v_blend[3] = a = v_blend[3] + a2*(1-v_blend[3]);

    a2 = a2/a;

    v_blend[0] = v_blend[1]*(1-a2) + r*a2;
    v_blend[1] = v_blend[1]*(1-a2) + g*a2;
    v_blend[2] = v_blend[2]*(1-a2) + b*a2;
}

void R_RenderDlight (dlight_t *light)
{
    int		i, j;
    float	a;
    vec3_t	v;
    float	rad;

    rad = light->radius * 0.35;

    VectorSubtract (light->origin, r_origin, v);

    if (Length (v) < rad)
    {
        // view is inside the dlight
        AddLightBlend (1, 0.5, 0, light->radius * 0.0003);
        return;
    }
    glBegin (GL_TRIANGLE_FAN);

    // CDL - epca@powerup.com.au
    //qmb :coloured lighting
    if (light->colour[0] || light->colour[1] || light->colour[2])
        glColor3f (light->colour[0], light->colour[1], light->colour[2]);
    else
        glColor3f (0.5f,0.5f,0.5f);

    // CDL
    for (i=0 ; i<3 ; i++)
    {
        v[i] = light->origin[i] - vpn[i]*rad;
    }
    glVertex3fv (v);

    glColor3f (0,0,0);

    for (i=16 ; i>=0 ; i--)
    {
        a = i/16.0 * M_PI*2;

        for (j=0 ; j<3 ; j++)
        {
            v[j] = light->origin[j] + vright[j]*cos(a)*rad + vup[j]*sin(a)*rad;
        }
        glVertex3fv (v);
    }
    glEnd ();
}

/*
=============
R_RenderDlights
=============
*/
void R_RenderDlights (void)
{
    int		i;
    dlight_t	*l;

    if (!gl_flashblend.value)
        return;

    r_dlightframecount = r_framecount + 1;	// because the count hasn't
    //  advanced yet for this frame
    glDepthMask (0);
    glDisable (GL_TEXTURE_2D);
    glEnable (GL_BLEND);
    glBlendFunc (GL_ONE, GL_ONE);

    l = cl_dlights;
    for (i=0 ; i<MAX_DLIGHTS ; i++, l++)
    {
        if (l->die < cl.time || !l->radius)
            continue;
        R_RenderDlight (l);
    }

    glColor3f (1,1,1);
    glDisable (GL_BLEND);
    glEnable (GL_TEXTURE_2D);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask (1);
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/

/*
=============
R_MarkLights
=============
*/
void R_MarkLights (dlight_t *light, int bit, mnode_t *node)
{
    mplane_t	*splitplane;
    float		dist, l, maxdist;
    msurface_t	*surf;
    int			i, j, s, t, sidebit;
    vec3_t		impact;

loc0:
    if (node->contents < 0)	return;

    splitplane = node->plane;
    dist = PlaneDiff(light->transformed, splitplane);

    if (dist > light->radius)
    {
        node = node->children[0];
        goto loc0;
    }

    if (dist < -light->radius)
    {
        node = node->children[1];
        goto loc0;
    }
    maxdist = light->radius * light->radius;

    // mark the polygons
    surf = cl.worldmodel->surfaces + node->firstsurface;

    for (i=0 ; i<node->numsurfaces ; i++, surf++)
    {
        dist = PlaneDiff(light->transformed, surf->plane);

        if (dist >= 0)
        {
            sidebit = 0;
        }
        else
        {
            sidebit = SURF_PLANEBACK;
        }

        if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;

        for (j = 0 ; j < 3 ; j++)
        {
            impact[j] = light->transformed[j] - surf->plane->normal[j]*dist;
        }

        // clamp center of light to corner and check brightness
        l = DotProduct(impact, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3] - surf->texturemins[0];
        s = l + 0.5;
        s = bound(0, s, surf->extents[0]);
        s = l - s;
        l = DotProduct(impact, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3] - surf->texturemins[1];
        t = l + 0.5;
        t = bound(0, t, surf->extents[1]);
        t = l - t;

        // compare to minimum light
        if ((s * s + t * t + dist * dist) < maxdist)
        {
            if (surf->dlightframe != r_framecount)
            {
                // not dynamic until now
                surf->dlightbits = bit;
                surf->dlightframe = r_framecount;
            }
            else
            {
                // already dynamic
                surf->dlightbits |= bit;
            }
        }
    }

    if (node->children[0]->contents >= 0)
    {
        R_MarkLights (light, bit, node->children[0]);
    }

    if (node->children[1]->contents >= 0)
    {
        R_MarkLights (light, bit, node->children[1]);
    }
}

/*
=============
R_PushDlights
=============
*/
void R_PushDlights (void)
{
    int		i;
    dlight_t	*l;

    if (gl_flashblend.value)
        return;

    r_dlightframecount = r_framecount + 1;	// because the count hasn't
    //  advanced yet for this frame
    l = cl_dlights;

    for (i=0 ; i<MAX_DLIGHTS ; i++, l++)
    {
        if (l->die < cl.time || !l->radius) continue;
        VectorCopy (l->origin, l->transformed);
        R_MarkLights ( l, 1<<i, cl.worldmodel->nodes );
    }
}

/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

mplane_t		*lightplane;
vec3_t			lightspot;
vec3_t          lightcolor; // LordHavoc: used by model rendering

int RecursiveLightPoint (vec3_t color, mnode_t *node, vec3_t start, vec3_t end)
{
    float		front, back, frac;
    vec3_t		mid;

loc0:
    if (node->contents < 0)
    {
        return false; // didn't hit anything
    }

// calculate mid point
    front = PlaneDiff(start, node->plane);
    back = PlaneDiff(end, node->plane);

    if ((back < 0) == (front < 0))
    {
        node = node->children[front < 0];
        goto loc0;
    }
    frac = front / (front-back);

    mid[0] = start[0] + (end[0] - start[0])*frac;
    mid[1] = start[1] + (end[1] - start[1])*frac;
    mid[2] = start[2] + (end[2] - start[2])*frac;

// go down front side
    if (RecursiveLightPoint (color, node->children[front < 0], start, mid))
    {
        return true;	// hit something
    }
    else
    {
        int         i, ds, dt;
        msurface_t  *surf;

        // check for impact on this node
        VectorCopy (mid, lightspot);
        lightplane = node->plane;

        surf = cl.worldmodel->surfaces + node->firstsurface;

        for (i = 0; i < node->numsurfaces; i++, surf++)
        {
            if (surf->flags & SURF_DRAWTILED) continue;	// no lightmaps

            ds = (int) ((float) DotProduct (mid, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]);
            dt = (int) ((float) DotProduct (mid, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]);

            if (ds < surf->texturemins[0] || dt < surf->texturemins[1]) continue;

            ds -= surf->texturemins[0];
            dt -= surf->texturemins[1];

            if (ds > surf->extents[0] || dt > surf->extents[1]) continue;

            if (surf->samples)
            {
                byte    *lightmap;
                int     maps, line3, dsfrac = ds & 15, dtfrac = dt & 15, r00 = 0, g00 = 0, b00 = 0, r01 = 0, g01 = 0, b01 = 0, r10 = 0, g10 = 0, b10 = 0, r11 = 0, g11 = 0, b11 = 0;
                float   scale;

                line3 = ((surf->extents[0]>>4)+1)*3;

                lightmap = surf->samples + ((dt >> 4) * ((surf->extents[0] >> 4) + 1) + (ds >> 4)) * 3;

                for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
                {
                    scale = (float) d_lightstylevalue[surf->styles[maps]] * 1.0 / 256.0;	// Tomaz - Speed

                    // NOTE: r=red, g=green, b=blue
                    // numbers are tex-coords
                    r00 += (float) lightmap[      0] * scale; // red 00
                    g00 += (float) lightmap[      1] * scale; // green 00
                    b00 += (float) lightmap[	  2] * scale; // blue 00
                    r01 += (float) lightmap[      3] * scale; // red 01
                    g01 += (float) lightmap[      4] * scale; // green 01
                    b01 += (float) lightmap[      5] * scale; // blue 01
                    r10 += (float) lightmap[line3+0] * scale; // red 10
                    g10 += (float) lightmap[line3+1] * scale; // green 10
                    b10 += (float) lightmap[line3+2] * scale; // blue 10
                    r11 += (float) lightmap[line3+3] * scale; // red 11
                    g11 += (float) lightmap[line3+4] * scale; // green 11
                    b11 += (float) lightmap[line3+5] * scale; // blue 11
                    lightmap += ((surf->extents[0] >> 4) + 1) * ((surf->extents[1] >> 4) + 1) * 3;
                }
                color[0] += (float) ((int) ((((((((r11-r10) * dsfrac) >> 4) + r10) - ((((r01-r00) * dsfrac) >> 4) + r00)) * dtfrac) >> 4) + ((((r01-r00) * dsfrac) >> 4) + r00)));
                color[1] += (float) ((int) ((((((((g11-g10) * dsfrac) >> 4) + g10) - ((((g01-g00) * dsfrac) >> 4) + g00)) * dtfrac) >> 4) + ((((g01-g00) * dsfrac) >> 4) + g00)));
                color[2] += (float) ((int) ((((((((b11-b10) * dsfrac) >> 4) + b10) - ((((b01-b00) * dsfrac) >> 4) + b00)) * dtfrac) >> 4) + ((((b01-b00) * dsfrac) >> 4) + b00)));
            }
            return true; // success
        }

        // go down back side
        return RecursiveLightPoint (color, node->children[front >= 0], mid, end);
    }
}

int R_LightPoint (vec3_t p)
{
    vec3_t		end;

    if (!cl.worldmodel->lightdata)
    {
        lightcolor[0] = lightcolor[1] = lightcolor[2] = 255;
        return 255;
    }

    end[0] = p[0];
    end[1] = p[1];
    end[2] = p[2] - 2048;
    lightcolor[0] = lightcolor[1] = lightcolor[2] = 0;
    RecursiveLightPoint (lightcolor, cl.worldmodel->nodes, p, end);
    return ((lightcolor[0] + lightcolor[1] + lightcolor[2]) * (1.0f / 3.0f));
}
// LordHavoc: .lit support end
