/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"

int		lightmap_bytes;		// 1, 2, or 4
int		lightmap_textures;

#define	BLOCK_WIDTH		128
#define	BLOCK_HEIGHT	128

#define	MAX_LIGHTMAPS	1024
int			active_lightmaps;

typedef struct glRect_s
{
    unsigned char l,t,w,h;
} glRect_t;

//temp lightmap
unsigned		blocklights[BLOCK_WIDTH*BLOCK_HEIGHT*3];
//qmb :coloured lights

glpoly_t	*lightmap_polys[MAX_LIGHTMAPS];
qboolean	lightmap_modified[MAX_LIGHTMAPS];
glRect_t	lightmap_rectchange[MAX_LIGHTMAPS];

int			allocated[MAX_LIGHTMAPS][BLOCK_WIDTH];

// the lightmap texture data needs to be kept in
// main memory so texsubimage can update properly
byte		lightmaps[4*MAX_LIGHTMAPS*BLOCK_WIDTH*BLOCK_HEIGHT];

//world texture chains
msurface_t  *skychain = NULL;
msurface_t  *waterchain = NULL;
msurface_t	*extrachain = NULL;
msurface_t	*outlinechain = NULL;

//qmb :detail texture
int		detailtexture;
int		detailtexture2;

//FIX ME
//remove soon...
//entity_t	*currententity;

void R_RenderDynamicLightmaps (msurface_t *fa);
void R_BuildLightMap (msurface_t *surf, byte *dest, int stride);

/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base)
{
    // MH - spelled relative correctly
    int		relative;
    int		count;

    // store the initial base in the event of an error
    texture_t *errorbase = base;

    if (base->alternate_anims)
    {
        base = base->alternate_anims;
    }

    if (!base->anim_total)
    {
        return base;
    }
    relative = (int)(cl.time*10) % base->anim_total;

    count = 0;

    while (base->anim_min > relative || base->anim_max <= relative)
    {
        base = base->anim_next;

        // if any error occurs, just put back the base texture
        if (!base) return errorbase;
        if (++count > 100) return errorbase;
    }
    return base;
}

/*
=============================================================

	BRUSH MODELS

=============================================================
*/

lp1DMTexFUNC qglMTexCoord1fARB = NULL;
lpMTexFUNC qglMTexCoord2fARB = NULL;
lpSelTexFUNC qglSelectTextureARB = NULL;

void GL_SelectTexture (GLenum target);

void GL_EnableTMU(int tmu)
{
    GL_SelectTexture(tmu);
    glEnable(GL_TEXTURE_2D);
}

void GL_DisableTMU(int tmu)
{
    GL_SelectTexture(tmu);
    glDisable(GL_TEXTURE_2D);
}

/*
=============================================================

	WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node, float *modelorg, int clipflags)
{
    int			c, side, sidebit;
    mplane_t	*plane;
    msurface_t	*surf, **mark;
    mleaf_t		*pleaf;
    double		dot;

    //make sure we are still inside the world
    if (node->contents == CONTENTS_SOLID) return;		// solid

    //is this node visable
    if (node->visframe != r_visframecount) return;

    // node culling as in software quake
    for (c = 0, plane = frustum ; c < 4 ; c++, plane++)
    {
        if (!(clipflags & (1 << c))) continue;	// don't need to clip against it

        if (BOX_ON_PLANE_SIDE(node->minmaxs, node->minmaxs + 3, plane) == 2) return;

        if (BOX_ON_PLANE_SIDE(node->minmaxs, node->minmaxs + 3, plane) == 1)
        {
            clipflags &= ~(1 << c);				// node is entirely on screen
        }
    }

// if a leaf node, draw stuff
    if (node->contents < 0)
    {
        pleaf = (mleaf_t *)node;
        mark = pleaf->firstmarksurface;
        c = pleaf->nummarksurfaces;

        if (c)
        {
            do
            {
                (*mark)->visframe = r_framecount;
                mark++;
            }
            while (--c);
        }

        // deal with model fragments in this leaf
        if (pleaf->efrags)
        {
            R_StoreEfrags (&pleaf->efrags);
        }
        return;
    }

// node is just a decision point, so go down the apropriate sides

// find which side of the node we are on
    plane = node->plane;
    dot = PlaneDiff(modelorg, plane);

    // find back facing verts
    side = (dot >= 0) ? 0 : 1;
    sidebit = (!side) ? 0 : SURF_PLANEBACK;

// recurse down the children, front side first
    R_RecursiveWorldNode (node->children[side], modelorg, clipflags);

// recurse down the back side
    if (r_outline.value)
    {
        R_RecursiveWorldNode (node->children[!side], modelorg, clipflags);
    }

// draw stuff
    c = node->numsurfaces;

    if (c)
    {
        for (surf = cl.worldmodel->surfaces + node->firstsurface ; c ; c--, surf++)
        {
            if (surf->visframe != r_framecount) continue;

            // which side are we on?
            if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;		// wrong side

            if (surf->flags & SURF_DRAWSKY)
            {
                surf->texturechain = skychain;
                skychain = surf;
            }
            else if (surf->flags & SURF_DRAWTURB)
            {
                surf->texturechain = waterchain;
                waterchain = surf;
            }
            else
            {
                //add chain for drawing.
                surf->texturechain = surf->texinfo->texture->texturechain;
                surf->texinfo->texture->texturechain = surf;

                //setup eyecandy chain
                if ((gl_detail.value&&gl_textureunits<4)
                || (surf->flags & SURF_UNDERWATER && gl_caustics.value)
                || (surf->texinfo->texture->gl_fullbright != 0 && gl_textureunits < 3)
                || (surf->flags & SURF_SHINY_METAL && gl_shiny.value)
                || (surf->flags & SURF_SHINY_GLASS && gl_shiny.value))
                {
                    surf->extra=extrachain;
                    extrachain = surf;
                }

                if (r_outline.value)
                {
                    surf->outline=outlinechain;
                    outlinechain = surf;
                }
            }
        }
    }

// recurse down the back side
    if (!r_outline.value)
    {
        R_RecursiveWorldNode (node->children[!side], modelorg, clipflags);
    }
}

/*
=================
R_DrawBrushModel
=================
*/
void R_DrawBrushModel (entity_t *e)
{
    int			k;
    vec3_t		mins, maxs;
    model_t		*clmodel;
    qboolean	rotated;
    int			i;
    int			lnum;
    vec3_t		dist;
    float       add;
    float		modelorg[3];

    clmodel = e->model;

    if (e->angles[0] || e->angles[1] || e->angles[2])
    {
        // model is rotated
        rotated = true;

        // mark as culled if the sphere is outside the frustum so that it wont have to be tested again
        if (R_CullSphere (e->origin, clmodel->radius)) return;
    }
    else
    {
        // model is normal
        rotated = false;

        VectorAdd (e->origin, clmodel->mins, mins);
        VectorAdd (e->origin, clmodel->maxs, maxs);

        // mark as culled if the box is outside the frustum so that it wont have to be tested again
        if (R_CullBox (mins, maxs)) return;
    }

//Lighting for model
//dynamic lights for non lightmaped bmodels
    if (clmodel->firstmodelsurface == 0)
    {
        R_LightPoint(e->origin);

        for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
        {
            if (cl_dlights[lnum].die >= cl.time)
            {
                VectorSubtract (e->origin, cl_dlights[lnum].origin, dist);
                add = cl_dlights[lnum].radius - Length(dist);

                if (add > 0)
                {
                    lightcolor[0] += add * cl_dlights[lnum].colour[0];
                    lightcolor[1] += add * cl_dlights[lnum].colour[1];
                    lightcolor[2] += add * cl_dlights[lnum].colour[2];
                }
            }
        }
        VectorScale(lightcolor, 1.0f / 100.0f, lightcolor);

        if (gl_ammoflash.value)
        {
            lightcolor[0] += sin(2 * cl.time * M_PI)/4;
            lightcolor[1] += sin(2 * cl.time * M_PI)/4;
            lightcolor[2] += sin(2 * cl.time * M_PI)/4;
        }
        glColor3fv (lightcolor);
    }
    else
    {
        glColor3f (1.0,1.0,1.0);
        memset (lightmap_polys, 0, sizeof(lightmap_polys));
    }
    VectorSubtract (r_refdef.vieworg, e->origin, modelorg);

    if (rotated)
    {
        vec3_t	temp;
        vec3_t	forward, right, up;

        VectorCopy (modelorg, temp);
        AngleVectors (e->angles, forward, right, up);
        modelorg[0] = DotProduct (temp, forward);
        modelorg[1] = -DotProduct (temp, right);
        modelorg[2] = DotProduct (temp, up);
    }

// calculate dynamic lighting for bmodel if it's not an
// instanced model
    // calculate dynamic lighting for bmodel if it's not an instanced model
    if (clmodel->firstmodelsurface != 0 && !gl_flashblend.value)
    {
        float matrix[16];

        // get the inverse transform for moving the light back to the same space as the model
        glPushMatrix ();
        glLoadIdentity ();
        glRotatef (-e->angles[2], 1, 0, 0);
        glRotatef (-e->angles[0], 0, 1, 0);
        glRotatef (-e->angles[1], 0, 0, 1);
        glTranslatef (-e->origin[0], -e->origin[1], -e->origin[2]);
        glGetFloatv (GL_MODELVIEW_MATRIX, matrix);
        glPopMatrix ();

        for (k = 0; k < MAX_DLIGHTS; k++)
        {
            if ((cl_dlights[k].die < cl.time) || (!cl_dlights[k].radius))
                continue;

            // move the light back to the same space as the model
            cl_dlights[k].transformed[0] = cl_dlights[k].origin[0] * matrix[0] + cl_dlights[k].origin[1] * matrix[4] + cl_dlights[k].origin[2] * matrix[8] + matrix[12];
            cl_dlights[k].transformed[1] = cl_dlights[k].origin[0] * matrix[1] + cl_dlights[k].origin[1] * matrix[5] + cl_dlights[k].origin[2] * matrix[9] + matrix[13];
            cl_dlights[k].transformed[2] = cl_dlights[k].origin[0] * matrix[2] + cl_dlights[k].origin[1] * matrix[6] + cl_dlights[k].origin[2] * matrix[10] + matrix[14];

            // light the model
            R_MarkLights (&cl_dlights[k], 1 << k, clmodel->nodes + clmodel->hulls[0].firstclipnode);
        }
    }

    /* erg what the hell is this ? */
    glPushMatrix ();
    e->angles[0] = -e->angles[0];	// stupid quake bug
    R_RotateForEntity (e);
    e->angles[0] = -e->angles[0];	// stupid quake bug

    R_DrawBrush (clmodel, &modelorg[0]);

    glPopMatrix ();
}

void R_DrawBrush (model_t *clmodel, float *modelorg)
{
    int			i;
    float		dot;
    mplane_t	*pplane;
    msurface_t	*surf;

    //if the chains havnt been made then make them
    for (surf = &clmodel->surfaces[clmodel->firstmodelsurface], i=0 ; i<clmodel->nummodelsurfaces ; i++, surf++)
    {
        // find which side of the node we are on
        pplane = surf->plane;
        dot = PlaneDiff(modelorg, pplane);

        // draw the polygon
        if (r_outline.value || ((surf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) || (!(surf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))
        {
            if (surf->flags & SURF_DRAWSKY)
            {
                surf->texturechain = skychain;
                skychain = surf;
            }
            else if (surf->flags & SURF_DRAWTURB)
            {
                surf->texturechain = waterchain;
                waterchain = surf;
            }
            else
            {
                //add chain for drawing.
                surf->texturechain = surf->texinfo->texture->texturechain;
                surf->texinfo->texture->texturechain = surf;

                //setup eyecandy chain
                if ((gl_detail.value&&gl_textureunits < 4)
                || (surf->flags & SURF_UNDERWATER && gl_caustics.value)
                || (surf->texinfo->texture->gl_fullbright != 0 && gl_textureunits < 3)
                || (surf->flags & SURF_SHINY_METAL && gl_shiny.value)
                || (surf->flags & SURF_SHINY_GLASS && gl_shiny.value))
                {
                    if (!R_Skybox)
                    {
                        surf->extra=extrachain;
                        extrachain = surf;
                    }
                }
                if (r_outline.value)
                {
                    surf->outline=outlinechain;
                    outlinechain = surf;
                }
            }
        }
    }

    if (gl_textureunits >= 4)
    {
        Surf_DrawTextureChainsFour (clmodel);
    }
    else
    {
        Surf_DrawTextureChainsTwo (clmodel);
    }
}

/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (void)
{
    entity_t	ent;
    float		modelorg[3];

    memset (&ent, 0, sizeof(ent));
    ent.model = cl.worldmodel;

    VectorCopy (r_refdef.vieworg, modelorg);

    glColor3f (1,1,1);
    memset (lightmap_polys, 0, sizeof(lightmap_polys));

    //draw the display list
    R_RecursiveWorldNode (cl.worldmodel->nodes, &modelorg[0], 15);

    //draw the world normally
    if (gl_textureunits >= 4)
    {
        Surf_DrawTextureChainsFour (cl.worldmodel);
    }
    else
    {
        Surf_DrawTextureChainsTwo (cl.worldmodel);
    }
}


/*
===============
R_MarkLeaves
===============
*/
void R_MarkLeaves (void)
{
    byte	*vis;
    mnode_t	*node;
    int		i;
    byte	solid[4096];

    if (r_oldviewleaf == r_viewleaf && !r_novis.value)
        return;

    r_visframecount++;
    r_oldviewleaf = r_viewleaf;

    if (r_novis.value)
    {
        vis = solid;
        memset (solid, 0xff, (cl.worldmodel->numleafs+7)>>3);
    }
    else vis = Mod_LeafPVS (r_viewleaf, cl.worldmodel);

    for (i=0 ; i<cl.worldmodel->numleafs ; i++)
    {
        if (vis[i>>3] & (1<<(i&7)))
        {
            node = (mnode_t *)&cl.worldmodel->leafs[i+1];
            do
            {
                if (node->visframe == r_visframecount)
                    break;
                node->visframe = r_visframecount;
                node = node->parent;
            }
            while (node);
        }
    }
}

/*
================
BuildSurfaceDisplayList
================
*/
void BuildSurfaceDisplayList (model_t *m, msurface_t *fa)
{
    int			i, lindex, lnumverts;
    medge_t		*pedges, *r_pedge;
    int			vertpage;
    float		*vec;
    float		s, t;
    glpoly_t	*poly;

// reconstruct the polygon
    pedges = m->edges;
    lnumverts = fa->numedges;
    vertpage = 0;

    //
    // draw texture
    //
    poly = Hunk_Alloc (sizeof(glpoly_t) + (lnumverts-4) * VERTEXSIZE*sizeof(float));
    poly->next = fa->polys;
    poly->flags = fa->flags;
    fa->polys = poly;
    poly->numverts = lnumverts;

    for (i=0 ; i<lnumverts ; i++)
    {
        lindex = m->surfedges[fa->firstedge + i];

        if (lindex > 0)
        {
            r_pedge = &pedges[lindex];
            vec = m->vertexes[r_pedge->v[0]].position;
        }
        else
        {
            r_pedge = &pedges[-lindex];
            vec = m->vertexes[r_pedge->v[1]].position;
        }
        s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
        s /= fa->texinfo->texture->width;

        t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
        t /= fa->texinfo->texture->height;

        VectorCopy (vec, poly->verts[i]);
        poly->verts[i][3] = s;
        poly->verts[i][4] = t;

        //
        // lightmap texture coordinates
        //
        s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
        s -= fa->texturemins[0];
        s += fa->light_s*16;
        s += 8;
        s /= BLOCK_WIDTH*16; //fa->texinfo->texture->width;

        t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
        t -= fa->texturemins[1];
        t += fa->light_t*16;
        t += 8;
        t /= BLOCK_HEIGHT*16; //fa->texinfo->texture->height;

        poly->verts[i][5] = s;
        poly->verts[i][6] = t;

        s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
        s /= 128;

        t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
        t /= 128;

        VectorCopy (vec, poly->verts[i]);
        poly->verts[i][7] = s;
        poly->verts[i][8] = t;
    }
    poly->numverts = lnumverts;

}

//==============================================================================================
//Lightmaps
//==============================================================================================
/*
=============================================================================

  LIGHTMAP ALLOCATION

=============================================================================
*/

// returns a texture number and the position inside it
int AllocBlock (int w, int h, int *x, int *y)
{
    int		i, j;
    int		best, best2;
    int		texnum;

    for (texnum=0 ; texnum<MAX_LIGHTMAPS ; texnum++)
    {
        best = BLOCK_HEIGHT;

        for (i=0 ; i<BLOCK_WIDTH-w ; i++)
        {
            best2 = 0;

            for (j=0 ; j<w ; j++)
            {
                if (allocated[texnum][i+j] >= best) break;

                if (allocated[texnum][i+j] > best2)
                {
                    best2 = allocated[texnum][i+j];
                }
            }

            if (j == w)
            {
                // this is a valid spot
                *x = i;
                *y = best = best2;
            }
        }

        if (best + h > BLOCK_HEIGHT) continue;

        for (i=0 ; i<w ; i++)
        {
            allocated[texnum][*x + i] = best + h;
        }
        return texnum;
    }
    Sys_Error ("AllocBlock: full");

    return 0;
}

/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (msurface_t *surf)
{
    int		smax, tmax;
    byte	*base;

    if (surf->flags & (SURF_DRAWSKY|SURF_DRAWTURB)) return;

    smax = (surf->extents[0]>>4)+1;
    tmax = (surf->extents[1]>>4)+1;

    surf->lightmaptexturenum = AllocBlock (smax, tmax, &surf->light_s, &surf->light_t);
    base = lightmaps + surf->lightmaptexturenum*lightmap_bytes*BLOCK_WIDTH*BLOCK_HEIGHT;
    base += (surf->light_t * BLOCK_WIDTH + surf->light_s) * lightmap_bytes;
    R_BuildLightMap (surf, base, BLOCK_WIDTH*lightmap_bytes);
}

void GL_UploadLightmap (void)
{
    int i;

    GL_SelectTexture(GL_TEXTURE0_ARB);

    //
    // upload all lightmaps that were filled
    //
    for (i=0 ; i<MAX_LIGHTMAPS ; i++)
    {
        if (!allocated[i][0])
            break;		// no more used
        lightmap_modified[i] = false;
        lightmap_rectchange[i].l = BLOCK_WIDTH;
        lightmap_rectchange[i].t = BLOCK_HEIGHT;
        lightmap_rectchange[i].w = 0;
        lightmap_rectchange[i].h = 0;
        glBindTexture(GL_TEXTURE_2D,lightmap_textures + i);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        glTexImage2D(GL_TEXTURE_2D, 0, lightmap_bytes, BLOCK_WIDTH, BLOCK_HEIGHT, 0, gl_lightmap_format, GL_UNSIGNED_BYTE, lightmaps+i*BLOCK_WIDTH*BLOCK_HEIGHT*lightmap_bytes);
    }
    GL_SelectTexture(GL_TEXTURE0_ARB);
}

/*
==================
GL_BuildLightmaps

Builds the lightmap texture
with all the surfaces from all brush models
==================
*/
void GL_BuildLightmaps (void)
{
    int		i,j;
    model_t	*m;
    extern vec3_t	lightcolor;

    memset (allocated, 0, sizeof(allocated));

    r_framecount = 1;		// no dlightcache

    if (!lightmap_textures)
    {
        lightmap_textures = texture_extension_number;
        texture_extension_number += MAX_LIGHTMAPS;
    }

    gl_lightmap_format = GL_RGBA;
    lightmap_bytes = 4;

    for (j=1 ; j<MAX_MODELS ; j++)
    {
        m = cl.model_precache[j];
        if (!m) break;
        if (m->name[0] == '*') continue;
        for (i=0 ; i<m->numsurfaces ; i++)
        {
            GL_CreateSurfaceLightmap (m->surfaces + i);
            if ( m->surfaces[i].flags & SURF_DRAWTURB ) continue;
            BuildSurfaceDisplayList (m, m->surfaces + i);
        }
    }
    GL_UploadLightmap();
}

/*
===============
R_AddDynamicLights
===============
*/
void R_AddDynamicLights (msurface_t *surf)
{
    int         lnum;
    int         sd, td;
    float       dist, rad, minlight;
    vec3_t      impact, local;
    int         s, t;
    int         i;
    int         smax, tmax;
    mtexinfo_t  *tex;
// LordHavoc: .lit support begin
    float		cred, cgreen, cblue, brightness;
    unsigned	*bl;
// LordHavoc: .lit support end

    smax = (surf->extents[0]>>4)+1;
    tmax = (surf->extents[1]>>4)+1;
    tex = surf->texinfo;

    for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
    {
        if ( !(surf->dlightbits & (1<<lnum) ) ) continue;       // not lit by this light

        rad = cl_dlights[lnum].radius;
        dist = PlaneDiff(cl_dlights[lnum].transformed, surf->plane);
        rad -= fabs(dist);
        minlight = cl_dlights[lnum].minlight;
        if (rad < minlight) continue;
        minlight = rad - minlight;

        for (i=0 ; i<3 ; i++)
        {
            impact[i] = cl_dlights[lnum].origin[i] - surf->plane->normal[i]*dist;
        }

        local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
        local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

        local[0] -= surf->texturemins[0];
        local[1] -= surf->texturemins[1];

// LordHavoc: .lit support begin
        bl = blocklights;
        cred = cl_dlights[lnum].colour[0] * 256.0f;
        cgreen = cl_dlights[lnum].colour[1] * 256.0f;
        cblue = cl_dlights[lnum].colour[2] * 256.0f;
// LordHavoc: .lit support end

        for (t = 0 ; t<tmax ; t++)
        {
            td = local[1] - t*16;
            if (td < 0)
                td = -td;
            for (s=0 ; s<smax ; s++)
            {
                sd = local[0] - s*16;
                if (sd < 0)
                    sd = -sd;
                if (sd > td)
                    dist = sd + (td>>1);
                else
                    dist = td + (sd>>1);
                if (dist < minlight)
                {
                    brightness = rad - dist;
                    bl[0] += (int) (brightness * cred);
                    bl[1] += (int) (brightness * cgreen);
                    bl[2] += (int) (brightness * cblue);
                }
                bl += 3;
            }
        }
    }
}

//directly changes values
__inline void RGBtoGrayscale(unsigned *rgb)
{
    unsigned value;
    //unsigned output;

    if (gl_sincity.value == 1)
    {
        value = rgb[0] * 0.299 + rgb[1] * 0.587 + rgb[2] * 0.114;

        rgb[0] = value;
        rgb[1] = value;
        rgb[2] = value;
    }
}

/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the 8.8 format in blocklights
===============
*/
void R_BuildLightMap (msurface_t *surf, byte *dest, int stride)
{
    int			smax, tmax;
    int			t;
    int			i, j, size;
    byte		*lightmap;
    unsigned	scale;
    int			maps;
    unsigned	*bl;

    surf->cached_dlight = (surf->dlightframe == r_framecount);

    smax = (surf->extents[0]>>4)+1;
    tmax = (surf->extents[1]>>4)+1;
    size = smax*tmax;
    lightmap = surf->samples;

// set to full bright if no light data
    if (!cl.worldmodel->lightdata)
    {
        bl = blocklights;
        for (i=0 ; i<size ; i++)
        {
            *bl++ = 255*256;//r
            *bl++ = 255*256;//g
            *bl++ = 255*256;//b
        }
        goto store;
    }

// clear to no light
    bl = blocklights;
    for (i=0 ; i<size ; i++)
    {
        *bl++ = 0;//r
        *bl++ = 0;//g
        *bl++ = 0;//b
    }

// add all the lightmaps
    if (lightmap)
        for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ;
                maps++)
        {
            unsigned *rgb;

            scale = d_lightstylevalue[surf->styles[maps]];
            surf->cached_light[maps] = scale;	// 8.8 fraction

            bl = blocklights;
            for (i=0 ; i<size ; i++)
            {
                rgb = bl;
                *bl++ += *lightmap++ * scale;//r
                *bl++ += *lightmap++ * scale;//g
                *bl++ += *lightmap++ * scale;//b

                RGBtoGrayscale(rgb);
            }

        }

    //qmb :overbright dynamic ligths only
    //so scale back the normal lightmap so it renders normally
    bl = blocklights;

    for (i=0; i<size*3 && gl_combine; i++)
    {
        *bl++ >>= 2;
    }

// add all the dynamic lights
    if (surf->dlightframe == r_framecount)
    {
        R_AddDynamicLights (surf);
    }

// bound, invert, and shift
store:
    stride -= (smax<<2);

    bl = blocklights;

    for (i=0 ; i<tmax ; i++, dest += stride)
    {
        for (j=0 ; j<smax ; j++)
        {
            //FIXME: should scale back so it remains the right colour
            // LordHavoc: .lit support begin
            t = bl[0] >> 7;
            if (t > 255) t = 255;
            *dest++ = t;//r
            t = bl[1] >> 7;
            if (t > 255) t = 255;
            *dest++ = t;//g
            t = bl[2] >> 7;
            if (t > 255) t = 255;
            *dest++ = t;//b
            bl = bl + 3;
            *dest++ = 255;
            // LordHavoc: .lit support end
        }
    }
}

/*
================
R_RenderDynamicLightmaps
Multitexture
================
*/
void R_RenderDynamicLightmaps (msurface_t *fa)
{
    byte		*base;
    int			maps;
    glRect_t    *theRect;
    int         smax, tmax;

    c_brush_polys++;

    if (fa->flags & ( SURF_DRAWSKY | SURF_DRAWTURB) )
        return;

    fa->polys->chain = lightmap_polys[fa->lightmaptexturenum];
    lightmap_polys[fa->lightmaptexturenum] = fa->polys;

    // check for lightmap modification
    for (maps = 0; maps < MAXLIGHTMAPS && fa->styles[maps] != 255; maps++)
    {
        if (d_lightstylevalue[fa->styles[maps]] != fa->cached_light[maps])
        {
            goto dynamic;
        }
    }

	// dynamic this frame
    if (fa->dlightframe == r_framecount || fa->cached_dlight)
    {
    // dynamic previously
dynamic:
        if (r_dynamic.value)
        {
            lightmap_modified[fa->lightmaptexturenum] = true;
            theRect = &lightmap_rectchange[fa->lightmaptexturenum];

            if (fa->light_t < theRect->t)
            {
                if (theRect->h)
                {
                    theRect->h += theRect->t - fa->light_t;
                }
                theRect->t = fa->light_t;
            }
            if (fa->light_s < theRect->l)
            {
                if (theRect->w)
                {
                    theRect->w += theRect->l - fa->light_s;
                }
                theRect->l = fa->light_s;
            }
            smax = (fa->extents[0]>>4)+1;
            tmax = (fa->extents[1]>>4)+1;

            if ((theRect->w + theRect->l) < (fa->light_s + smax))
            {
                theRect->w = (fa->light_s-theRect->l)+smax;
            }

            if ((theRect->h + theRect->t) < (fa->light_t + tmax))
            {
                theRect->h = (fa->light_t-theRect->t)+tmax;
            }
            base = lightmaps + fa->lightmaptexturenum*lightmap_bytes*BLOCK_WIDTH*BLOCK_HEIGHT;
            base += fa->light_t * BLOCK_WIDTH * lightmap_bytes + fa->light_s * lightmap_bytes;
            R_BuildLightMap (fa, base, BLOCK_WIDTH*lightmap_bytes);
        }
    }
}
