/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"

extern int image_width;
extern int image_height;

/*
=========================================================

TARGA LOADING

=========================================================
*/

/*
=============
TARGA LOADING
=============
*/
#define TGA_MAXCOLORS 16384

/* Defitions for image types. */
#define TGA_Null		0	/* no image image_rgba */
#define TGA_Map			1	/* Uncompressed, color-mapped images. */
#define TGA_RGB			2	/* Uncompressed, RGB images. */
#define TGA_Mono		3	/* Uncompressed, black and white images. */
#define TGA_RLEMap		9	/* Runlength encoded color-mapped images. */
#define TGA_RLERGB		10	/* Runlength encoded RGB images. */
#define TGA_RLEMono		11	/* Compressed, black and white images. */
#define TGA_CompMap		32	/* Compressed color-mapped image_rgba, using Huffman, Delta, and runlength encoding. */
#define TGA_CompMap4	33	/* Compressed color-mapped image_rgba, using Huffman, Delta, and runlength encoding.  4-pass quadtree-type process. */

/* Defitions for interleave flag. */
#define TGA_IL_None		0	/* non-interleaved. */
#define TGA_IL_Two		1	/* two-way (even/odd) interleaving */
#define TGA_IL_Four		2	/* four way interleaving */
#define TGA_IL_Reserved	3	/* reserved */

/* Defitions for origin flag */
#define TGA_O_UPPER		0	/* Origin in lower left-hand corner. */
#define TGA_O_LOWER		1	/* Origin in upper left-hand corner. */

typedef struct _TargaHeader
{
    unsigned char 	id_length, colormap_type, image_type;
    unsigned short	colormap_index, colormap_length;
    unsigned char	colormap_size;
    unsigned short	x_origin, y_origin, width, height;
    unsigned char	pixel_size, attributes;
} TargaHeader;

int fgetLittleShort (FILE *f)
{
    byte	b1, b2;

    b1 = fgetc(f);
    b2 = fgetc(f);

    return (short)(b1 + b2*256);
}

int fgetLittleLong (FILE *f)
{
    byte	b1, b2, b3, b4;

    b1 = fgetc(f);
    b2 = fgetc(f);
    b3 = fgetc(f);
    b4 = fgetc(f);

    return b1 + (b2<<8) + (b3<<16) + (b4<<24);
}

/*
========
LoadTGA
========
*/
byte *LoadTGA (FILE *f, char *name)
{
    int			w, h, x, y, realrow, truerow, baserow, i, temp1, temp2, pixel_size, map_idx;
    int			RLE_count, RLE_flag, size, interleave, origin;
    qboolean	mapped, rlencoded;
    byte		*image_rgba, *dst, r, g, b, a, j, k, l, *ColorMap;
    TargaHeader	header;

    header.id_length = fgetc (f);
    header.colormap_type = fgetc (f);
    header.image_type = fgetc (f);

    header.colormap_index = fgetLittleShort (f);
    header.colormap_length = fgetLittleShort (f);
    header.colormap_size = fgetc (f);
    header.x_origin = fgetLittleShort (f);
    header.y_origin = fgetLittleShort (f);
    header.width = fgetLittleShort (f);
    header.height = fgetLittleShort (f);
    header.pixel_size = fgetc (f);
    header.attributes = fgetc (f);

    if (!f && COM_FOpenFile(name, &f)) return NULL;

    if (header.id_length != 0)
        fseek (f, header.id_length, SEEK_CUR);

    /* validate TGA type */
    switch (header.image_type)
    {
    case TGA_Map:
    case TGA_RGB:
    case TGA_Mono:
    case TGA_RLEMap:
    case TGA_RLERGB:
    case TGA_RLEMono:
        break;

    default:
        Con_Printf ("%s : Only type 1 (map), 2 (RGB), 3 (mono), 9 (RLEmap), 10 (RLERGB), 11 (RLEmono) TGA images supported\n", name);
        fclose (f);
        return NULL;
        break;
    }

    /* validate color depth */
    switch (header.pixel_size)
    {
    case 8:
    case 15:
    case 16:
    case 24:
    case 32:
        break;

    default:
        Con_Printf ("%s : Only 8, 15, 16, 24 or 32 bit images (with colormaps) supported\n", name);
        fclose (f);
        return NULL;
        break;
    }
    r = g = b = a = l = 0;

    /* if required, read the color map information. */
    ColorMap = NULL;
    mapped = (header.image_type == TGA_Map || header.image_type == TGA_RLEMap) && header.colormap_type == 1;

    if (mapped)
    {
        /* validate colormap size */
        switch (header.colormap_size)
        {
        case 8:
        case 15:
        case 16:
        case 32:
        case 24:
            break;

        default:
            Con_Printf ("%s : Only 8, 15, 16, 24 or 32 bit colormaps supported\n", name);
            fclose (f);
            return NULL;
            break;
        }
        temp1 = header.colormap_index;
        temp2 = header.colormap_length;

        if ((temp1 + temp2 + 1) >= TGA_MAXCOLORS)
        {
            fclose (f);
            return NULL;
        }
        ColorMap = malloc (TGA_MAXCOLORS * 4);
        map_idx = 0;

        for (i = temp1 ; i < temp1 + temp2 ; ++i, map_idx += 4)
        {
            /* read appropriate number of bytes, break into rgb & put in map. */
            switch (header.colormap_size)
            {
            case 8:	/* grey scale, read and triplicate. */
                r = g = b = getc (f);
                a = 255;
                break;

            case 15:	/* 5 bits each of red green and blue. */
                /* watch byte order. */
                j = getc (f);
                k = getc (f);
                l = ((unsigned int) k << 8) + j;
                r = (byte) ( ((k & 0x7C) >> 2) << 3 );
                g = (byte) ( (((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3 );
                b = (byte) ( (j & 0x1F) << 3 );
                a = 255;
                break;

            case 16:	/* 5 bits each of red green and blue, 1 alpha bit. */
                /* watch byte order. */
                j = getc (f);
                k = getc (f);
                l = ((unsigned int) k << 8) + j;
                r = (byte) ( ((k & 0x7C) >> 2) << 3 );
                g = (byte) ( (((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3 );
                b = (byte) ( (j & 0x1F) << 3 );
                a = (k & 0x80) ? 255 : 0;
                break;

            case 24:	/* 8 bits each of blue, green and red. */
                b = getc (f);
                g = getc (f);
                r = getc (f);
                a = 255;
                l = 0;
                break;

            case 32:	/* 8 bits each of blue, green, red and alpha. */
                b = getc (f);
                g = getc (f);
                r = getc (f);
                a = getc (f);
                l = 0;
                break;
            }
            ColorMap[map_idx+0] = r;
            ColorMap[map_idx+1] = g;
            ColorMap[map_idx+2] = b;
            ColorMap[map_idx+3] = a;
        }
    }

    /* check run-length encoding. */
    rlencoded = (header.image_type == TGA_RLEMap || header.image_type == TGA_RLERGB || header.image_type == TGA_RLEMono);
    RLE_count = RLE_flag = 0;

    image_width = w = header.width;
    image_height = h = header.height;

    size = w * h * 4;
    image_rgba = calloc (size, 1);

    /* read the Targa file body and convert to portable format. */
    pixel_size = header.pixel_size;
    origin = (header.attributes & 0x20) >> 5;
    interleave = (header.attributes & 0xC0) >> 6;
    truerow = 0;
    baserow = 0;

    for (y=0 ; y<h ; y++)
    {
        realrow = truerow;

        if (origin == TGA_O_UPPER)
        {
            realrow = h - realrow - 1;
        }
        dst = image_rgba + realrow * w * 4;

        for (x=0 ; x<w ; x++)
        {
            /* check if run length encoded. */
            if (rlencoded)
            {
                if (!RLE_count)
                {
                    /* have to restart run. */
                    i = getc (f);

                    RLE_flag = (i & 0x80);

                    if (!RLE_flag)	// stream of unencoded pixels
                    {
                        RLE_count = i + 1;
                    }
                    else		// single pixel replicated
                    {
                        RLE_count = i - 127;
                    }

                    /* decrement count & get pixel. */
                    --RLE_count;
                }
                else
                {
                    /* have already read count & (at least) first pixel. */
                    --RLE_count;

                    if (RLE_flag)
                    {
                        /* replicated pixels. */
                        goto PixEncode;
                    }
                }
            }

            /* read appropriate number of bytes, break into RGB. */
            switch (pixel_size)
            {
            case 8:	/* grey scale, read and triplicate. */
                r = g = b = l = getc (f);
                a = 255;
                break;

            case 15: /* 5 bits each of red green and blue. */
                /* watch byte order. */
                j = getc (f);
                k = getc (f);
                l = ((unsigned int) k << 8) + j;
                r = (byte) ( ((k & 0x7C) >> 2) << 3 );
                g = (byte) ( (((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3 );
                b = (byte) ( (j & 0x1F) << 3 );
                a = 255;
                break;

            case 16: /* 5 bits each of red green and blue, 1 alpha bit. */
                /* watch byte order. */
                j = getc (f);
                k = getc (f);
                l = ((unsigned int) k << 8) + j;
                r = (byte) ( ((k & 0x7C) >> 2) << 3 );
                g = (byte) ( (((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3 );
                b = (byte) ( (j & 0x1F) << 3 );
                a = (k & 0x80) ? 255 : 0;
                break;

            case 24: /* 8 bits each of blue, green and red. */
                b = getc (f);
                g = getc (f);
                r = getc (f);
                a = 255;
                l = 0;
                break;

            case 32: /* 8 bits each of blue, green, red and alpha. */
                b = getc (f);
                g = getc (f);
                r = getc (f);
                a = getc (f);
                l = 0;
                break;

            default:
                fclose (f);
                Con_Printf ("%s : Illegal pixel_size '%d'\n", name, pixel_size);
                if (image_rgba)
                {
                    free (image_rgba);
                }

                if (mapped)
                {
                    free (ColorMap);
                }
                return NULL;
                break;
            }

PixEncode:
            if (mapped)
            {
                map_idx = l * 4;
                *dst++ = ColorMap[map_idx+0];
                *dst++ = ColorMap[map_idx+1];
                *dst++ = ColorMap[map_idx+2];
                *dst++ = ColorMap[map_idx+3];
            }
            else
            {
                *dst++ = r;
                *dst++ = g;
                *dst++ = b;
                *dst++ = a;
            }
        }

        if (interleave == TGA_IL_Four)
        {
            truerow += 4;
        }
        else if (interleave == TGA_IL_Two)
        {
            truerow += 2;
        }
        else
        {
            truerow++;
        }

        if (truerow >= h)
        {
            truerow = ++baserow;
        }
    }

    if (mapped)
    {
        free (ColorMap);
    }
    fclose (f);

    return image_rgba;
}


